from .api_key import set_key  # type: ignore
from .core import Kit, async_pnwkit, pnwkit  # type: ignore

__version__ = "1.3.1"

# shortcuts for pnwkit.xxx syntax as opposed to pnwkit.pnwkit.xxx
alliance_query = pnwkit.alliance_query
color_query = pnwkit.color_query
nation_query = pnwkit.nation_query
trade_price_query = pnwkit.trade_price_query
trade_query = pnwkit.trade_query
treasure_query = pnwkit.treasure_query
async_war_query = pnwkit.war_query

async_alliance_query = async_pnwkit.alliance_query
async_color_query = async_pnwkit.color_query
async_nation_query = async_pnwkit.nation_query
async_trade_price_query = async_pnwkit.trade_price_query
async_trade_query = async_pnwkit.trade_query
async_treasure_query = async_pnwkit.treasure_query
async_war_query = async_pnwkit.war_query



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

from .async_ import AsyncKit
from .sync import SyncKit

if TYPE_CHECKING:
    from typing import Any, Optional, Union


class Kit:
    """Class for the creation of a kit.

    For simple creation of a SyncKit or AsyncKit

    Returns
    -------
    Union[AsyncKit, SyncKit]
        SyncKit if ``async_`` is ``False``, AsyncKit if ``async_`` is ``True``.
    """

    @overload
    def __new__(
        cls,
        api_key: Optional[str] = ...,
        *,
        async_: Literal[False] = False,
        **kwargs: Any
    ) -> SyncKit:
        ...

    @overload
    def __new__(
        cls,
        api_key: Optional[str] = ...,
        *,
        async_: Literal[True] = True,
        **kwargs: Any
    ) -> AsyncKit:
        ...

    def __new__(
        cls, api_key: Optional[str] = None, *, async_: bool = False, **kwargs: Any
    ) -> Union[AsyncKit, SyncKit]:
        if "async" in kwargs:
            async_ = kwargs["async"]
        if async_:
            return AsyncKit(api_key=api_key, **kwargs)
        return SyncKit(api_key=api_key, **kwargs)


pnwkit: SyncKit = Kit()

async_pnwkit: AsyncKit = Kit(async_=True)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, overload

from .api_key import API_KEY

__all__ = ("KitBase",)

if TYPE_CHECKING:
    from typing import (
        Any,
        List,
        Literal,
        Mapping,
        MutableMapping,
        Optional,
        Sequence,
        Tuple,
        Union,
    )

    from .data import *
    from .paginator import *

    SubQuery = Mapping[str, Union[str, "SubQuery", Sequence[Union[str, "SubQuery"]]]]


class KitBase(metaclass=ABCMeta):
    def __init__(self, api_key: Optional[str] = None, **kwargs: Any) -> None:
        self.api_key: str = api_key or API_KEY

    def graphql_url(self) -> str:
        return f"https://api.politicsandwar.com/graphql?api_key={self.api_key}"

    def set_key(self, api_key: str) -> None:
        """Sets the API key for this instance.

        Parameters
        ----------
        api_key : str
            A Politics and War API Key.
        """
        self.api_key: str = api_key

    @classmethod
    def _format_sub_query(
        cls,
        query: SubQuery,
    ) -> str:
        key, query_string = list(query.items())[0]
        if not isinstance(query_string, str):
            query_arguments: List[Any] = []
            for value in query_string:
                if isinstance(value, str):
                    query_arguments.append(value)
                else:
                    query_arguments.append(cls._format_sub_query(value))
            query_string = " ".join(query_arguments)

        return f"{key}{{{query_string}}}"

    @classmethod
    def _format_query(
        cls,
        endpoint: str,
        params: MutableMapping[str, Any],
        query: Union[
            str,
            Sequence[Union[str, Mapping[str, Union[str, Mapping[str, Any]]]]],
        ],
        is_paginator: bool,
    ) -> str:
        if isinstance(params, str):
            params_string = params
        else:
            params_string = ",".join(
                name
                + ":"
                + (
                    f'"{value}"'
                    if isinstance(value, str)
                    else str(value).lower()
                    if isinstance(value, bool)
                    else str(value)
                )
                for name, value in params.items()
            )
        if isinstance(query, str):
            query_string = query
        else:
            query_arguments: List[Any] = []
            for value in query:
                if isinstance(value, str):
                    query_arguments.append(value)
                else:
                    query_arguments.append(cls._format_sub_query(value))
            query_string = " ".join(query_arguments)
        if is_paginator:
            if params:
                return f"{{{endpoint}({params_string}){{paginatorInfo{{count currentPage firstItem hasMorePages lastItem lastPage perPage total}}data{{{query_string}}}}}}}"
            return f"{{{endpoint}{{paginatorInfo{{count currentPage firstItem hasMorePages lastItem lastPage perPage total}}data{{{query_string}}}}}}}"
        if params_string:
            return f"{{{endpoint}({params_string}){{{query_string}}}}}"
        return f"{{{endpoint}{{{query_string}}}}}"

    @overload
    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False],
        **kwargs: Any,
    ) -> Tuple[Alliance, ...]:
        ...

    @overload
    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True],
        **kwargs: Any,
    ) -> AlliancePaginator:
        ...

    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **kwargs: Any,
    ) -> Union[Tuple[Alliance, ...], AlliancePaginator]:
        """Makes a query to the alliances endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Union[Tuple[Alliance], AlliancePaginator]
            A tuple of Alliances representing the data retrived. If ``paginator`` is True then will return an AlliancePaginator.
        """
        ...

    @abstractmethod
    def color_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Color, ...]:
        """Makes a query to the colors endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Tuple[Color]
            A tuple of Colors representing the data retrived.
        """
        ...

    @overload
    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = False,
        **kwargs: Any,
    ) -> Tuple[Nation, ...]:
        ...

    @overload
    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = True,
        **kwargs: Any,
    ) -> NationPaginator:
        ...

    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **kwargs: Any,
    ) -> Union[Tuple[Nation, ...], NationPaginator]:
        """Makes a query to the nations endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Union[Tuple[Nation], NationPaginator]
            A tuple of Nations representing the data retrived. If ``paginator`` is True then will return a NationPaginator.
        """
        ...

    @abstractmethod
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Trade, ...]:
        """Makes a query to the trades endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Tuple[Trade]
            A tuple of Trades representing the data retrived.
        """
        ...

    @abstractmethod
    def trade_price_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Tradeprice, ...]:
        """Makes a query to the tradeprices endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Tuple[Tradeprice]
            A tuple of Tradeprices representing the data retrived.
        """
        ...

    @abstractmethod
    def treasure_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Treasure, ...]:
        """Makes a query to the treasures endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Tuple[Treasure]
            A tuple of Treasures representing the data retrived.
        """
        ...

    @abstractmethod
    def war_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[War, ...]:
        """Makes a query to the wars endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        arg : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the data to retrieve.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated with arg to form the string of data to retrieve.
        **kwargs: Any
            If params is falsy, then extra kwargs will be used as params.

        Returns
        -------
        Tuple[War]
            A tuple of Wars representing the data retrived.
        """
        ...



API_KEY = None 


def set_key(api_key: str) -> None:
    """Sets the default API key and the API key on the :data:`pnwkit` and :data:`async_pnwkit` instances.

    Parameters
    ----------
    api_key : str
        A Politics and War API Key.
    """
    from .core import async_pnwkit, pnwkit

    global API_KEY
    API_KEY = api_key
    async_pnwkit.set_key(api_key)
    pnwkit.set_key(api_key)



from __future__ import annotations

from typing import Any, Dict, Mapping


class DumpData:
    id: str

    __slots__ = ()

    def __init__(self, data: Mapping[str, Any]) -> None:
        for key, value in data.items():
            attr = __builtins__[self.__annotations__[key]]
            object.__setattr__(self, key, attr(value))

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delattr__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __setitem__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delitem__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __getitem__(self, name: str) -> Any:
        try:
            return self.__getattribute__(name)
        except AttributeError:
            raise KeyError(name)

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {int(self)}"

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self.__getattribute__(key)
        except AttributeError:
            return default

    def __int__(self) -> int:
        raise NotImplementedError  # needs to be implemented in subclasses

    def to_dict(self) -> Dict[str, Any]:
        """
        Get a dict representation of this data.

        Returns
        -------
        Dict[str, Any]
            A dictionary of the data on the object.
        """
        return self.__dict__


class DumpAlliance(DumpData):
    """Represents an alliance.

    Attributes
    ----------
    alliance_id: :class:`str`
        The alliance ID.
    date_created: :class:`str`
        The date the alliance was created.
    name: :class:`str`
        The alliance name.
    acronym: :class:`str`
        The alliance acronym.
    color: :class:`str`
        The color of the alliance.
    continent: :class:`str`
        The continent the alliance is in.
    discord_server: :class:`str`
        The invite link to the alliance's Discord server.
    score: :class:`float`
        The current score of the alliance.
    flag_url: :class:`str`
        The URL to the alliance's flag.
    """

    alliance_id: str
    date_created: str
    name: str
    acronym: str
    color: str
    continent: str
    discord_server: str
    score: float
    flag_url: str

    __slots__ = (
        "alliance_id",
        "date_created",
        "name",
        "acronym",
        "color",
        "continent",
        "discord_server",
        "score",
        "flag_url",
    )

    def __int__(self) -> int:
        return int(self.get("alliance_id", -1))


class DumpCity(DumpData):
    """Represents a city.

    Attributes
    ----------
    city_id: :class:`str`
        The city ID.
    nation_id: :class:`str`
        The nation ID of the city's nation.
    date_created: :class:`str`
        The date the city was created.
    name: :class:`str`
        The name of the city.
    capital: :class:`bool`
        Whether the city is the capital of its nation or not.
    infrastructure: :class:`float`
        The amount of infrastructure in the city.
    maxinfra: :class:`float`
        The maximum amount of infrastructure in the city.
    land: :class:`float`
        The amount of land in the city.
    oil_power_plants: :class:`int`
        The amount of Oil Power Plants in the city.
    wind_power_plants: :class:`int`
        The amount of Wind Power Plants in the city.
    coal_power_plants: :class:`int`
        The amount of Coal Power Plants in the city.
    nuclear_power_plants: :class:`int`
        The amount of Nuclear Power Plants in the city.
    coal_mines: :class:`int`
        The amount of Coal Mines in the city.
    oil_wells: :class:`int`
        The amount of Oil Wells in the city.
    uranium_mines: :class:`int`
        The amount of Uranium Mines in the city.
    iron_mines: :class:`int`
        The amount of Iron Mines in the city.
    lead_mines: :class:`int`
        The amount of Lead Mines in the city.
    bauxite_mines: :class:`int`
        The amount of Bauxite Mines in the city.
    farms: :class:`int`
        The amount of Farms in the city.
    police_stations: :class:`int`
        The amount of Police Stations in the city.
    hospitals: :class:`int`
        The amount of Hospitals in the city.
    recycling_centers: :class:`int`
        The amount of Recycling Centers in the city.
    subway: :class:`int`
        The amount of Subways in the city.
    supermarkets: :class:`int`
        The amount of Supermarkets in the city.
    banks: :class:`int`
        The amount of Banks in the city.
    shopping_malls: :class:`int`
        The amount of Shopping Malls in the city.
    stadiums: :class:`int`
        The amount of Stadiums in the city.
    oil_refineries: :class:`int`
        The amount of Oil Refineries in the city.
    aluminum_refineries: :class:`int`
        The amount of Aluminum Refineries in the city.
    steel_mills: :class:`int`
        The amount of Steel Mill in the city.
    munitions_factories: :class:`int`
        The amount of Munitions Factories in the city.
    barracks: :class:`int`
        The amount of Barracks in the city.
    factories: :class:`int`
        The amount of Factories in the city.
    hangars: :class:`int`
        The amount of Hangars in the city.
    drydocks: :class:`int`
        The amount of Drydocks in the city.
    last_nuke_date: :class:`str`
        The date the city was last nuked.
    """

    city_id: str
    nation_id: str
    date_created: str
    name: str
    capital: bool
    infrastructure: float
    maxinfra: float
    land: float
    oil_power_plants: int
    wind_power_plants: int
    coal_power_plants: int
    nuclear_power_plants: int
    coal_mines: int
    oil_wells: int
    uranium_mines: int
    iron_mines: int
    lead_mines: int
    bauxite_mines: int
    farms: int
    police_stations: int
    hospitals: int
    recycling_centers: int
    subway: int
    supermarkets: int
    banks: int
    shopping_malls: int
    stadiums: int
    oil_refineries: int
    aluminum_refineries: int
    steel_mills: int
    munitions_factories: int
    barracks: int
    factories: int
    hangars: int
    drydocks: int
    last_nuke_date: str

    __slots__ = (
        "city_id",
        "nation_id",
        "date_created",
        "name",
        "capital",
        "infrastructure",
        "maxinfra",
        "land",
        "oil_power_plants",
        "wind_power_plants",
        "coal_power_plants",
        "nuclear_power_plants",
        "coal_mines",
        "oil_wells",
        "uranium_mines",
        "iron_mines",
        "lead_mines",
        "bauxite_mines",
        "farms",
        "police_stations",
        "hospitals",
        "recycling_centers",
        "subway",
        "supermarkets",
        "banks",
        "shopping_malls",
        "stadiums",
        "oil_refineries",
        "aluminum_refineries",
        "steel_mills",
        "munitions_factories",
        "barracks",
        "factories",
        "hangars",
        "drydocks",
        "last_nuke_date",
    )

    def __int__(self) -> int:
        return int(self.get("city_id", -1))


class DumpNation(DumpData):
    """Represents a nation.

    Attributes
    ----------
    nation_id: :class:`str`
        The nation ID.
    nation_name: :class:`str`
        The nation name.
    leader_name: :class:`str`
        The nation's leader name.
    date_created: :class:`str`
        When the nation was created.
    continent: :class:`str`
        The nation's continent.
    latitude: :class:`float`
        The nation's latitude.
    longitude: :class:`float`
        The nation's longitude.
    leader_title: :class:`str`
        The nation's leader title.
    nation_title: :class:`str`
        The nation's title.
    score: :class:`float`
        The nation's score.
    population: :class:`int`
        The nation's population.
    flag_url: :class:`str`
        The URL of the nation's flag.
    color: :class:`str`
        The nation's color.
    beige_turns_remaining: :class:`int`
        The number of turns the nation has on beige.
    portrait_url: :class:`str`
        The URL of the nation's portrait.
    cities: :class:`int`
        The nation's cities.
    gdp: :class:`int`
        The nation's GDP.
    currency: :class:`str`
        The nation's currency.
    wars_won: :class:`int`
        The number of wars the nation has won.
    wars_lost: :class:`int`
        The number of wars the nation has lost.
    alliance: :class:`str`
        The nation's alliance.
    alliance_id: :class:`str`
        The nation's alliance ID.
    alliance_position: :class:`int`
        The nation's alliance position.
    soldiers: :class:`int`
        The amount of soldiers the nation has.
    tanks: :class:`int`
        The amount of tanks the nation has.
    aircraft: :class:`int`
        The amount of aircraft the nation has.
    ships: :class:`int`
        The amount of ships the nation has.
    missiles: :class:`int`
        The amount of missiles the nation has.
    nukes: :class:`int`
        The amount of nukes the nation has.
    domestic_policy: :class:`str`
        The nation's domestic policy.
    war_policy: :class:`str`
        The nation's war policy.
    projects: :class:`int`
        The amount of projects the nation has.
    ironworks_np: :class:`bool`
        Whether the nation has the Ironworks project or not.
    bauxiteworks_np: :class:`bool`
        Whether the nation has the Bauxiteworks project or not.
    arms_stockpile_np: :class:`bool`
        Whether the nation has the Arms Stockpile project or not.
    emergency_gasoline_reserve_np: :class:`bool`
        Whether the nation has the Emergency Gasoline Reserve project or not.
    mass_irrigation_np: :class:`bool`
        Whether the nation has the Mass Irrigation project or not.
    international_trade_center_np: :class:`bool`
        Whether the nation has the International Trade Center project or not.
    missile_launch_pad_np: :class:`bool`
        Whether the nation has the Missile Launch Pad project or not.
    nuclear_research_facility_np: :class:`bool`
        Whether the nation has the Nuclear Research Facility project or not.
    iron_dome_np: :class:`bool`
        Whether the nation has the Iron Dome project or not.
    vital_defense_system_np: :class:`bool`
        Whether the nation has the Vital Defense System project or not.
    intelligence_agency_np: :class:`bool`
        Whether the nation has the Intelligence Agency project or not.
    center_for_civil_engineering_np: :class:`bool`
        Whether the nation has the Center for Civil Engineering project or not.
    propaganda_bureau_np: :class:`bool`
        Whether the nation has the Propaganda Bureau project or not.
    uranium_enrichment_program_np: :class:`bool`
        Whether the nation has the Uranium Enrichment Program project or not.
    urban_planning_np: :class:`bool`
        Whether the nation has the Urban Planning project or not.
    advanced_urban_planning_np: :class:`bool`
        Whether the nation has the Advanced Urban Planning project or not.
    space_program_np: :class:`bool`
        Whether the nation has the Space Program project or not.
    moon_landing_np: :class:`bool`
        Whether the nation has the Moon Landing project or not.
    spy_satellite_np: :class:`bool`
        Whether the nation has the Spy Satellite project or not.
    pirate_economy_np: :class:`bool`
        Whether the nation has the Pirate Economy project or not.
    recycling_initiative_np: :class:`bool`
        Whether the nation has the Recycling Initiative project or not.
    telecommunications_satellite_np: :class:`bool`
        Whether the nation has the Telecommunication Satellite project or not.
    green_technologies_np: :class:`bool`
        Whether the nation has the Green Technologies project or not.
    clinical_research_center_np: :class:`bool`
        Whether the nation has the Clinical Research Center project or not.
    specialized_police_training_program_np: :class:`bool`
        Whether the nation has the Specialized Police Training Program project or not.
    arable_land_agency_np: :class:`bool`
        Whether the nation has the Arable Land Agency project or not.
    advanced_engineering_corps_np: :class:`bool`
        Whether the nation has the Advanced Engineering Corps project or not.
    vm_turns: :class:`int`
        The number of turns the nation has on Vacation Mode.
    """

    nation_id: str
    nation_name: str
    leader_name: str
    date_created: str
    continent: str
    latitude: float
    longitude: float
    leader_title: str
    nation_title: str
    score: float
    population: int
    flag_url: str
    color: str
    beige_turns_remaining: int
    portrait_url: str
    cities: int
    gdp: int
    currency: str
    wars_won: int
    wars_lost: int
    alliance: str
    alliance_id: str
    alliance_position: int
    soldiers: int
    tanks: int
    aircraft: int
    ships: int
    missiles: int
    nukes: int
    domestic_policy: str
    war_policy: str
    projects: int
    ironworks_np: bool
    bauxiteworks_np: bool
    arms_stockpile_np: bool
    emergency_gasoline_reserve_np: bool
    mass_irrigation_np: bool
    international_trade_center_np: bool
    missile_launch_pad_np: bool
    nuclear_research_facility_np: bool
    iron_dome_np: bool
    vital_defense_system_np: bool
    intelligence_agency_np: bool
    center_for_civil_engineering_np: bool
    propaganda_bureau_np: bool
    uranium_enrichment_program_np: bool
    urban_planning_np: bool
    advanced_urban_planning_np: bool
    space_program_np: bool
    moon_landing_np: bool
    spy_satellite_np: bool
    pirate_economy_np: bool
    recycling_initiative_np: bool
    telecommunications_satellite_np: bool
    green_technologies_np: bool
    clinical_research_center_np: bool
    specialized_police_training_program_np: bool
    arable_land_agency_np: bool
    advanced_engineering_corps_np: bool
    vm_turns: int

    __slots__ = (
        "nation_id",
        "nation_name",
        "leader_name",
        "date_created",
        "continent",
        "latitude",
        "longitude",
        "leader_title",
        "nation_title",
        "score",
        "population",
        "flag_url",
        "color",
        "beige_turns_remaining",
        "portrait_url",
        "cities",
        "gdp",
        "currency",
        "wars_won",
        "wars_lost",
        "alliance",
        "alliance_id",
        "alliance_position",
        "soldiers",
        "tanks",
        "aircraft",
        "ships",
        "missiles",
        "nukes",
        "domestic_policy",
        "war_policy",
        "projects",
        "ironworks_np",
        "bauxiteworks_np",
        "arms_stockpile_np",
        "emergency_gasoline_reserve_np",
        "mass_irrigation_np",
        "international_trade_center_np",
        "missile_launch_pad_np",
        "nuclear_research_facility_np",
        "iron_dome_np",
        "vital_defense_system_np",
        "intelligence_agency_np",
        "center_for_civil_engineering_np",
        "propaganda_bureau_np",
        "uranium_enrichment_program_np",
        "urban_planning_np",
        "advanced_urban_planning_np",
        "space_program_np",
        "moon_landing_np",
        "spy_satellite_np",
        "pirate_economy_np",
        "recycling_initiative_np",
        "telecommunications_satellite_np",
        "green_technologies_np",
        "clinical_research_center_np",
        "specialized_police_training_program_np",
        "arable_land_agency_np",
        "advanced_engineering_corps_np",
        "vm_turns",
    )

    def __int__(self) -> int:
        return int(self.get("nation_id", -1))


class DumpTrade(DumpData):
    """Represents a trade

    Attributes
    ----------
    trade_id: :class:`str`
        The trade ID.
    date_created: :class:`str`
        The date the trade was created.
    offerer_nation_id: :class:`str`
        The nation ID of the offerer.
    receiver_nation_id: :class:`str`
        The nation ID of the receiver.
    offer_type: :class:`str`
        The type of offer.
    buy_or_sell: :class:`str`
        Whether the offer was a `buy` or `sell` offer.
    resource: :class:`str`
        The resource being offered.
    quantity: :class:`int`
        The amount being offered.
    price: :class:`int`
        The price per unit of the offer.
    accepted: :class:`bool`
        Whether or not the offer has been accepted.
    original_trade_id: :class:`str`
        The trade ID of the original trade.
    date_accepted: :class:`str`
        The date the offer was accepted.
    """

    trade_id: str
    date_created: str
    offerer_nation_id: str
    receiver_nation_id: str
    offer_type: str
    buy_or_sell: str
    resource: str
    quantity: int
    price: int
    accepted: bool
    original_trade_id: str
    date_accepted: str

    __slots__ = (
        "trade_id",
        "date_created",
        "offerer_nation_id",
        "receiver_nation_id",
        "offer_type",
        "buy_or_sell",
        "resource",
        "quantity",
        "price",
        "accepted",
        "original_trade_id",
        "date_accepted",
    )

    def __int__(self) -> int:
        return int(self.get("trade_id", -1))


class DumpWar(DumpData):
    """Represents a war.

    Attributes
    ----------
    war_id: :class:`str`
        The war ID.
    date_declared: :class:`str`
        The date the war was declared.
    aggressor_nation_id: :class:`str`
        The attacker ID.
    defender_nation_id: :class:`str`
        The defender ID.
    aggressor_alliance_name: :class:`str`
        The aggressor's alliance name.
    aggressor_alliance_id: :class:`str`
        The aggressor's alliance ID.
    aggressor_alliance_position: :class:`int`
        The position of the aggressor's alliance.
    defender_alliance_name: :class:`str`
        The defender's alliance name.
    defender_alliance_id: :class:`str`
        The defender's alliance ID.
    defender_alliance_position: :class:`int`
        The position of the defender's alliance.
    aggressor_offering_peace: :class:`bool`
        Whether the aggressor is offering peace or not.
    defender_offering_peace: :class:`bool`
        Whether the defender is offering peace or not.
    reason: :class:`str`
        The war reason.
    ground_control: :class:`str`
        The ID of the nation that has Ground Control.
    air_superiority: :class:`str`
        The ID of the nation that has Air Superiority.
    blockade: :class:`str`
        The ID of the nation that has a Naval Blockade.
    turnsleft: :class:`int`
        The turns left in the war.
    aggressor_resistance: :class:`int`
        The aggressor's resistance.
    defender_resistance: :class:`int`
        The defender's resistance.
    war_type: :class:`str`
        The war type.
    aggressor_war_policy: :class:`str`
        The aggressor's war policy.
    defender_war_policy: :class:`str`
        The defender's war policy.
    att_attacks: :class:`int`
        The number of attacks the aggressor has made.
    def_attacks: :class:`int`
        The number of attacks the defender has made.
    att_gas_used: :class:`float`
        The amount of gasoline used by the attacker.
    def_gas_used: :class:`float`
        The amount of gasoline used by the defender.
    att_mun_used: :class:`float`
        The amount of munitions used by the attacker.
    def_mun_used: :class:`float`
        The amount of munitions used by the defender.
    att_alum_used: :class:`int`
        The amount of aluminum used by the attacker.
    def_alum_used: :class:`int`
        The amount of aluminum used by the defender.
    att_steel_used: :class:`int`
        The amount of steel used by the attacker.
    def_steel_used: :class:`int`
        The amount of steel used by the defender.
    att_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the attacker.
    def_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the defender.
    att_money_looted: :class:`float`
        The amount of money looted by the attacker.
    def_money_looted: :class:`float`
        The amount of money looted by the attacker.
    att_soldiers_killed: :class:`int`
        The amount of soldiers killed by the attacker.
    def_soldiers_killed: :class:`int`
        The amount of soldiers killed by the defender.
    att_tanks_killed: :class:`int`
        The amount of tanks killed by the attacker.
    def_tanks_killed: :class:`int`
        The amount of tanks killed by the defender.
    att_aircraft_killed: :class:`int`
        The amount of aircraft killed by the attacker.
    def_aircraft_killed: :class:`int`
        The amount of aircraft killed by the defender.
    att_ships_killed: :class:`int`
        The amount of ships killed by the attacker.
    def_ships_killed: :class:`int`
        The amount of ships killed by the defender.
    att_missiles_used: :class:`int`
        The amount of missiles used by the attacker.
    def_missiles_used: :class:`int`
        The amount of missiles killed by the defender.
    att_nukes_used: :class:`int`
        The amount of nukes used by the attacker.
    def_nukes_used: :class:`int`
        The amount of nukes killed by the defender.
    att_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the attacker.
    def_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the defender.
    """

    war_id: str
    date_declared: str
    aggressor_nation_id: str
    defender_nation_id: str
    aggressor_alliance_name: str
    aggressor_alliance_id: str
    aggressor_alliance_position: int
    defender_alliance_name: str
    defender_alliance_id: str
    defender_alliance_position: int
    aggressor_offering_peace: bool
    defender_offering_peace: bool
    reason: str
    ground_control: str
    air_superiority: str
    blockade: str
    turns_left: int
    aggressor_resistance: int
    defender_resistance: int
    war_type: str
    aggressor_war_policy: str
    defender_war_policy: str
    att_attacks: int
    def_attacks: int
    att_gas_used: float
    def_gas_used: float
    att_mun_used: float
    def_mun_used: float
    att_alum_used: int
    def_alum_used: int
    att_steel_used: int
    def_steel_used: int
    att_infra_destroyed: float
    def_infra_destroyed: float
    att_money_looted: float
    def_money_looted: float
    att_soldiers_killed: int
    def_soldiers_killed: int
    att_tanks_killed: int
    def_tanks_killed: int
    att_aircraft_killed: int
    def_aircraft_killed: int
    att_ships_killed: int
    def_ships_killed: int
    att_missiles_used: int
    def_missiles_used: int
    att_nukes_used: int
    def_nukes_used: int
    att_infra_destroyed_value: float
    def_infra_destroyed_value: float

    __slots__ = (
        "war_id",
        "date_declared",
        "aggressor_nation_id",
        "defender_nation_id",
        "aggressor_alliance_name",
        "aggressor_alliance_id",
        "aggressor_alliance_position",
        "defender_alliance_name",
        "defender_alliance_id",
        "defender_alliance_position",
        "aggressor_offering_peace",
        "defender_offering_peace",
        "reason",
        "ground_control",
        "air_superiority",
        "blockade",
        "turns_left",
        "aggressor_resistance",
        "defender_resistance",
        "war_type",
        "aggressor_war_policy",
        "defender_war_policy",
        "att_attacks",
        "def_attacks",
        "att_gas_used",
        "def_gas_used",
        "att_mun_used",
        "def_mun_used",
        "att_alum_used",
        "def_alum_used",
        "att_steel_used",
        "def_steel_used",
        "att_infra_destroyed",
        "def_infra_destroyed",
        "att_money_looted",
        "def_money_looted",
        "att_soldiers_killed",
        "def_soldiers_killed",
        "att_tanks_killed",
        "def_tanks_killed",
        "att_aircraft_killed",
        "def_aircraft_killed",
        "att_ships_killed",
        "def_ships_killed",
        "att_missiles_used",
        "def_missiles_used",
        "att_nukes_used",
        "def_nukes_used",
        "att_infra_destroyed_value",
        "def_infra_destroyed_value",
    )

    def __int__(self) -> int:
        return int(self.get("war_id", -1))


_TYPE_MAP = {
    "alliances": DumpAlliance,
    "cities": DumpCity,
    "nations": DumpNation,
    "trades": DumpTrade,
    "wars": DumpWar,
}



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Sequence

from .data import Alliance, Data, Nation, PaginatorInfo

__all__ = ("AlliancePaginator", "NationPaginator")

if TYPE_CHECKING:
    from typing import Any, Sequence


class Paginator(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, data: Sequence[Data], paginator_info: PaginatorInfo) -> None:
        self.__dict__["data"] = data
        self.__dict__["paginator_info"] = paginator_info

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delattr__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __setitem__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delitem__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __getitem__(self, name: str) -> Any:
        return self.__getattribute__(name)

    def __repr__(self) -> str:
        return type(self).__name__


class AlliancePaginator(Paginator):
    def __init__(self, data: Sequence[Alliance], paginator_info: PaginatorInfo) -> None:
        self.__dict__["data"] = data
        self.__dict__["paginator_info"] = paginator_info


class NationPaginator(Paginator):
    def __init__(self, data: Sequence[Nation], paginator_info: PaginatorInfo) -> None:
        self.__dict__["data"] = data
        self.__dict__["paginator_info"] = paginator_info



from __future__ import annotations

from typing import TYPE_CHECKING

__all__ = (
    "Alliance",
    "Bankrec",
    "City",
    "Color",
    "Nation",
    "Trade",
    "Tradeprice",
    "Treasure",
    "War",
    "WarAttack",
    "PaginatorInfo",
)

if TYPE_CHECKING:
    from typing import Any, Dict, Mapping, Sequence, Type


class Data:
    id: str
    __slots__ = ()

    def __init__(self, data: Mapping[str, Any]) -> None:
        for key, value in data.items():
            if isinstance(value, dict):
                attr = _ATTRIBUTE_MAP[key]
                object.__setattr__(self, key, attr(value))  # type: ignore
            elif isinstance(value, list):
                attr = _ATTRIBUTE_MAP[key]
                object.__setattr__(self, key, tuple(attr(i) for i in value))  # type: ignore
            else:
                object.__setattr__(self, key, value)

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delattr__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __setitem__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delitem__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __getitem__(self, name: str) -> Any:
        try:
            return self.__getattribute__(name)
        except AttributeError:
            raise KeyError(name)

    def __int__(self) -> int:
        return int(self.get("id", -1))

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {int(self)}"

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self.__getattribute__(key)
        except AttributeError:
            return default

    def to_dict(self) -> Dict[str, Any]:
        """
        Get a dict representation of this data.

        Returns
        -------
        Dict[str, Any]
            A dictionary of the data on the object.
        """
        return self.__dict__


class Alliance(Data):
    """Represents an alliance.

    Attributes
    ----------
    id: :class:`str`
        The alliance ID.
    name: :class:`str`
        The alliance name.
    acronym: :class:`str`
        The alliance acronym.
    score: :class:`float`
        The alliance score.
    color: :class:`str`
        The alliance color.
    nations: Sequence[:class:`Nation`]
        The alliance's members.
    acceptmem: :class:`bool`
        Whether the alliance is accepting members or not.
    flag: :class:`str`
        The alliance flag URL.
    forumlink: :class:`str`
        The alliance forum link.
    irclink: :class:`str`
        The alliance Discord link.
    bankrecs: Sequence[:class:`Bankrec`]
        Sequence containing the alliance's bank records. Will be `None` if the querying key cannot view the alliance bank.
    taxrecs: Sequence[:class:`Bankrec`]
        Sequence containing the alliance's tax records. Will be `None` if the querying key cannot view the alliance bank.
    money: :class:`float`
        The current amount of money in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    coal: :class:`float`
        The current amount of coal in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    oil: :class:`float`
        The current amount of oil in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    uranium: :class:`float`
        The current amount of uranium in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    iron: :class:`float`
        The current amount of iron in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    bauxite: :class:`float`
        The current amount of bauxite in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    lead: :class:`float`
        The current amount of lead in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    gasoline: :class:`float`
        The current amount of gasoline in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    munitions: :class:`float`
        The current amount of munitions in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    steel: :class:`float`
        The current amount of steel in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    aluminum: :class:`float`
        The current amount of aluminum in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    food: :class:`float`
        The current amount of food in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    """

    id: str
    name: str
    acronym: str
    score: float
    color: str
    nations: Sequence[Nation]
    acceptmem: bool
    flag: str
    forumlink: str
    irclink: str
    bankrecs: Sequence[Bankrec]
    taxrecs: Sequence[Bankrec]
    money: float
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float

    __slots__ = (
        "id",
        "name",
        "acronym",
        "score",
        "color",
        "nations",
        "acceptmem",
        "flag",
        "forumlink",
        "irclink",
        "bankrecs",
        "taxrecs",
        "money",
        "coal",
        "oil",
        "uranium",
        "iron",
        "bauxite",
        "lead",
        "gasoline",
        "munitions",
        "steel",
        "aluminum",
        "food",
    )


class Bankrec(Data):
    """Represents a bank record.

    Attributes
    ----------
    id: :class:`str`
        The bank record ID.
    date: :class:`str`
        The date the transaction was completed.
    sid: :class:`str`
        The ID of the sender.
    stype: :class:`int`
        The type of the sender. 1 corresponds to a :class:`Nation` and 2 corresponds to a :class:`Alliance`.
    rid: :class:`str`
        The ID of the receiver.
    rtype: :class:`int`
        The type of the receiver. 1 corresponds to a :class:`Nation` and 2 corresponds to a :class:`Alliance`
    pid: :class:`str`
        The nation ID of the banker.
    note: :class:`str`
        The transaction note.
    money: :class:`float`
        The money sent in the transaction.
    coal: :class:`float`
        The coal sent in the transaction.
    oil: :class:`float`
        The oil sent in the transaction.
    uranium: :class:`float`
        The uranium sent in the transaction.
    iron: :class:`float`
        The iron sent in the transaction.
    bauxite: :class:`float`
        The bauxite sent in the transaction.
    lead: :class:`float`
        The lead sent in the transaction.
    gasoline: :class:`float`
        The gasoline sent in the transaction.
    munitions: :class:`float`
        The munitions sent in the transaction.
    steel: :class:`float`
        The steel sent in the transaction.
    aluminum: :class:`float`
        The aluminum sent in the transaction.
    food: :class:`float`
        The food sent in the transaction.
    tax_id: :class:`str`
        The ID of the tax bracker.
    """

    id: str
    date: str
    sid: str
    stype: int
    rid: str
    rtype: int
    pid: str
    note: str
    money: float
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float
    tax_id: str

    __slots__ = (
        "id",
        "date",
        "sid",
        "stype",
        "rid",
        "rtype",
        "pid",
        "note",
        "money",
        "coal",
        "oil",
        "uranium",
        "iron",
        "bauxite",
        "lead",
        "gasoline",
        "munitions",
        "steel",
        "aluminum",
        "food",
        "tax_id",
    )


class City(Data):
    """Represents a city.

    Attributes
    ----------
    id: :class:`str`
        The city ID.
    name: :class:`str`
        The city name.
    date: :class:`str`
        The date the city was created.
    infrastructure: :class:`float`
        The amount of infrastructure in the city.
    land: :class:`float`
        The amount of land in the city.
    powered: :class:`bool`
        Whether the city is powered or not.
    oilpower: :class:`int`
        The amount of Oil Power Plants in the city.
    windpower: :class:`int`
        The amount of Wind Power Plants in the city.
    coalpower: :class:`int`
        The amount of Coal Power Plants in the city.
    nuclearpower: :class:`int`
        The amount of Nuclear Power Plants in the city.
    coalmine: :class:`int`
        The amount of Coal Mines in the city.
    oilwell: :class:`int`
        The amount of Oil Wells in the city.
    uramine: :class:`int`
        The amount of Uranium Mines in the city.
    barracks: :class:`int`
        The amount of Barracks in the city.
    farm: :class:`int`
        The amount of Farms in the city.
    policestation: :class:`int`
        The amount of Police Stations in the city.
    hospital: :class:`int`
        The amount of Hospitals in the city.
    recyclingcenter: :class:`int`
        The amount of Recycling Centers in the city.
    subway: :class:`int`
        The amount of Subways in the city.
    supermarket: :class:`int`
        The amount of Supermarkets in the city.
    bank: :class:`int`
        The amount of Banks in the city.
    mall: :class:`int`
        The amount of Shopping Malls in the city.
    stadium: :class:`int`
        The amount of Stadiums in the city.
    leadmine: :class:`int`
        The amount of Lead Mines in the city.
    ironmine: :class:`int`
        The amount of Iron Mines in the city.
    bauxitemine: :class:`int`
        The amount of Bauxite Mines in the city.
    gasrefinery: :class:`int`
        The amount of Oil Refineries in the city.
    aluminumrefinery: :class:`int`
        The amount of Aluminum Refineries in the city.
    steelmill: :class:`int`
        The amount of Steel Mill in the city.
    munitionsfactory: :class:`int`
        The amount of Munitions Factories in the city.
    factory: :class:`int`
        The amount of Factories in the city.
    airforcebase: :class:`int`
        The amount of Hangars in the city.
    drydock: :class:`int`
        The amount of Drydocks in the city.
    """

    id: str
    name: str
    date: str
    infrastructure: float
    land: float
    powered: bool
    oilpower: int
    windpower: int
    coalpower: int
    nuclearpower: int
    coalmine: int
    oilwell: int
    uramine: int
    barracks: int
    farm: int
    policestation: int
    hospital: int
    recyclingcenter: int
    subway: int
    supermarket: int
    bank: int
    mall: int
    stadium: int
    leadmine: int
    ironmine: int
    bauxitemine: int
    gasrefinery: int
    aluminumrefinery: int
    steelmill: int
    munitionsfactory: int
    factory: int
    airforcebase: int
    drydock: int

    __slots__ = (
        "id",
        "name",
        "date",
        "infrastructure",
        "land",
        "powered",
        "oilpower",
        "windpower",
        "coalpower",
        "nuclearpower",
        "coalmine",
        "oilwell",
        "uramine",
        "barracks",
        "farm",
        "policestation",
        "hospital",
        "recyclingcenter",
        "subway",
        "supermarket",
        "bank",
        "mall",
        "stadium",
        "leadmine",
        "ironmine",
        "bauxitemine",
        "gasrefinery",
        "aluminumrefinery",
        "steelmill",
        "munitionsfactory",
        "factory",
        "airforcebase",
        "drydock",
    )


class Color(Data):
    """Represents a color bloc.

    Attributes
    ----------
    color: :class:`str`
        The color.
    bloc_name: :class:`str`
        The color bloc name.
    turn_bonus: :class:`int`
        The color turn bonus.
    """

    color: str
    bloc_name: str
    turn_bonus: int

    __slots__ = ("color", "bloc_name", "turn_bonus")

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {self.bloc_name if 'bloc_name' in self.__dict__ else -1}"


class Nation(Data):
    """Represents a nation.

    Attributes
    ----------
    id: :class:`str`
        The nation ID.
    alliance_id: :class:`str`
        The nation's alliance ID.
    alliance_position: :class:`str`
        The nation's alliance position.
    alliance: :class:`Alliance`
        The nation's alliance.
    nation_name: :class:`str`
        The nation name.
    leader_name: :class:`str`
        The nation's leader name.
    continent: :class:`str`
        The nation's continent.
    warpolicy: :class:`str`
        The nation's war policy.
    dompolicy: :class:`str`
        The nation's domestic policy.
    color: :class:`str`
        The nation's color
    num_cities: :class:`int`
        The number of cities the nation has.
    cities: Sequence[:class:`City`]
        The nation's cities.
    score: :class:`float`
        The nation's score.
    population: :class:`int`
        The nation's population.
    flag: :class:`str`
        The nation's flag.
    vmode: :class:`int`
        The number of turns the nation has in Vacation Mode.
    beigeturns: :class:`int`
        The number of turns the nation has on Beige.
    espionage_available: :class:`bool`
        Whether the nation has espionage available or not.
    last_active: :class:`str`
        When the nation was last active.
    date: :class:`str`
        When the nation was created.
    soldiers: :class:`int`
        The amount of soldiers the nation has.
    tanks: :class:`int`
        The amount of tanks the nation has.
    aircraft: :class:`int`
        The amount of aircraft the nation has.
    ships: :class:`int`
        The amount of ships the nation has.
    missiles: :class:`int`
        The amount of missiles the nation has.
    nukes: :class:`int`
        The amount of nukes the nation has.
    treasures: Sequence[:class:`Treasure`]
        The treasures a nation has.
    offensive_wars: Sequence[:class:`War`]
        The offensive wars the nation is involved in.
    defensive_wars: Sequence[:class:`War`]
        The defensive wars the nation is involved in.
    sent_bankrecs: Sequence[:class:`Bankrec`]
        The sent bank records of the nation.
    received_bankrecs: Sequence[:class:`Bankrec`]
        The received bank records of the nation.
    money: :class:`float`
        The amount of money the nation has. Will return `None` if the querying key cannot see the nation's bank.
    coal: :class:`float`
        The amount of coal the nation has. Will return `None` if the querying key cannot see the nation's bank.
    oil: :class:`float`
        The amount of oil the nation has. Will return `None` if the querying key cannot see the nation's bank.
    uranium: :class:`float`
        The amount of uranium the nation has. Will return `None` if the querying key cannot see the nation's bank.
    iron: :class:`float`
        The amount of iron the nation has. Will return `None` if the querying key cannot see the nation's bank.
    bauxite: :class:`float`
        The amount of bauxite the nation has. Will return `None` if the querying key cannot see the nation's bank.
    lead: :class:`float`
        The amount of lead the nation has. Will return `None` if the querying key cannot see the nation's bank.
    gasoline: :class:`float`
        The amount of gasoline the nation has. Will return `None` if the querying key cannot see the nation's bank.
    munitions: :class:`float`
        The amount of munitions the nation has. Will return `None` if the querying key cannot see the nation's bank.
    steel: :class:`float`
        The amount of steel the nation has. Will return `None` if the querying key cannot see the nation's bank.
    aluminum: :class:`float`
        The amount of aluminum the nation has. Will return `None` if the querying key cannot see the nation's bank.
    food: :class:`float`
        The amount of food the nation has. Will return `None` if the querying key cannot see the nation's bank.
    projects: :class:`int`
        The number of projects the nation has.
    ironw: :class:`bool`
        Whether the nation has the Ironworks project or not.
    bauxitew: :class:`bool`
        Whether the nation has the Bauxiteworks project or not.
    armss: :class:`bool`
        Whether the nation has the Arms Stockpile project or not.
    egr: :class:`bool`
        Whether the nation has the Emergency Gasoline Reserve project or not.
    massirr: :class:`bool`
        Whether the nation has the Mass Irrigation project or not.
    itc: :class:`bool`
        Whether the nation has the International Trade Center project or not.
    mlp: :class:`bool`
        Whether the nation has the Missile Launch Pad project or not.
    nrf: :class:`bool`
        Whether the nation has the Nuclear Research Facility project or not.
    irond: :class:`bool`
        Whether the nation has the Iron Dome project or not.
    vds: :class:`bool`
        Whether the nation has the Vital Defense System project or not.
    cia: :class:`bool`
        Whether the nation has the Intelligence Agency project or not.
    cfce: :class:`bool`
        Whether the nation has the Center for Civil Engineering project or not.
    propb: :class:`bool`
        Whether the nation has the Propaganda Bureau project or not.
    uap: :class:`bool`
        Whether the nation has the Uranium Enrichment Program project or not.
    city_planning: :class:`bool`
        Whether the nation has the Urban Planning project or not.
    adv_city_planning: :class:`bool`
        Whether the nation has the Advanced Urban Planning project or not.
    space_program: :class:`bool`
        Whether the nation has the Space Program project or not.
    spy_satellite: :class:`bool`
        Whether the nation has the Spy Satellite project or not.
    moon_landing: :class:`bool`
        Whether the nation has the Moon Landing project or not.
    pirate_economy: :class:`bool`
        Whether the nation has the Pirate Economy project or not.
    recycling_initiative: :class:`bool`
        Whether the nation has the Recycling Initiative project or not.
    telecom_satellite: :class:`bool`
        Whether the nation has the Telecommunication Satellite project or not.
    green_tech: :class:`bool`
        Whether the nation has the Green Technologies project or not.
    arable_land_agency: :class:`bool`
        Whether the nation has the Arable Land Agency project or not.
    clinical_research_center: :class:`bool`
        Whether the nation has the Clinical Research Center project or not.
    specialized_police_training: :class:`bool`
        Whether the nation has the Specialized Police Training Program project or not.
    adv_engineering_corps: :class:`bool`
        Whether the nation has the Advanced Engineering Corps project or not.
    """

    id: str
    alliance_id: str
    alliance_position: str
    alliance: Alliance
    nation_name: str
    leader_name: str
    continent: str
    warpolicy: str
    dompolicy: str
    color: str
    num_cities: int
    cities: Sequence[City]
    score: float
    population: int
    flag: str
    vmode: int
    beigeturns: int
    espionage_available: bool
    last_active: str
    date: str
    soldiers: int
    tanks: int
    aircraft: int
    ships: int
    missiles: int
    nukes: int
    treasures: Sequence[Treasure]
    offensive_wars: Sequence[War]
    defensive_wars: Sequence[War]
    sent_bankrecs: Sequence[Bankrec]
    received_bankrecs: Sequence[Bankrec]
    money: float
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float
    projects: int
    ironw: bool
    bauxitew: bool
    armss: bool
    egr: bool
    massirr: bool
    itc: bool
    mlp: bool
    nrf: bool
    irond: bool
    vds: bool
    cia: bool
    cfce: bool
    propb: bool
    uap: bool
    city_planning: bool
    adv_city_planning: bool
    space_program: bool
    spy_satellite: bool
    moon_landing: bool
    pirate_economy: bool
    recycling_initiative: bool
    telecom_satellite: bool
    green_tech: bool
    arable_land_agency: bool
    clinical_research_center: bool
    specialized_police_training: bool
    adv_engineering_corps: bool

    __slots__ = (
        "id",
        "alliance_id",
        "alliance_position",
        "alliance",
        "nation_name",
        "leader_name",
        "continent",
        "warpolicy",
        "dompolicy",
        "color",
        "num_cities",
        "cities",
        "score",
        "population",
        "flag",
        "vmode",
        "beigeturns",
        "espionage_available",
        "last_active",
        "date",
        "soldiers",
        "tanks",
        "aircraft",
        "ships",
        "missiles",
        "nukes",
        "treasures",
        "offensive_wars",
        "defensive_wars",
        "sent_bankrecs",
        "received_bankrecs",
        "money",
        "coal",
        "oil",
        "uranium",
        "iron",
        "bauxite",
        "lead",
        "gasoline",
        "munitions",
        "steel",
        "aluminum",
        "food",
        "projects",
        "ironw",
        "bauxitew",
        "armss",
        "egr",
        "massirr",
        "itc",
        "mlp",
        "nrf",
        "irond",
        "vds",
        "cia",
        "cfce",
        "propb",
        "uap",
        "city_planning",
        "adv_city_planning",
        "space_program",
        "spy_satellite",
        "moon_landing",
        "pirate_economy",
        "recycling_initiative",
        "telecom_satellite",
        "green_tech",
        "arable_land_agency",
        "clinical_research_center",
        "specialized_police_training",
        "adv_engineering_corps",
    )


class Trade(Data):
    """Represents a trade

    Attributes
    ----------
    id: :class:`str`
        The trade ID.
    date: :class:`str`
        The date the trade took place.
    sid: :class:`str`
        The sender ID.
    rid: :class:`str`
        The receiver ID.
    sender: :class:`Nation`
        The sender's nation.
    receiver: :class:`Nation`
        The receiver's nation.
    offer_resource: :class:`str`
        The resource being offered.
    offer_amount: :class:`int`
        The amount being offered.
    buy_or_sell: :class:`str`
        Whether the offer was a `buy` or `sell` offer.
    total: :class:`int`
        The price per unit of the offer.
    accepted: :class:`bool`
        Whether or not the offer has been accepted.
    date_accepted: :class:`str`
        The date the offer was accepted.
    """

    id: str
    date: str
    sid: str
    rid: str
    sender: Nation
    receiver: Nation
    offer_resource: str
    offer_amount: int
    buy_or_sell: str
    total: int
    accepted: bool
    date_accepted: str

    __slots__ = (
        "id",
        "date",
        "sid",
        "rid",
        "sender",
        "receiver",
        "offer_resource",
        "offer_amount",
        "buy_or_sell",
        "total",
        "accepted",
        "date_accepted",
    )


class Tradeprice(Data):
    """Represents a tradeprice.

    Attributes
    ----------
    id: :class:`str`
        The tradeprice ID.
    date: :class:`str`
        The date of the tradeprice.
    coal: :class:`float`
        The price of coal.
    oil: :class:`float`
        The price of oil.
    uranium: :class:`float`
        The price of uranium.
    iron: :class:`float`
        The price of iron.
    bauxite: :class:`float`
        The price of bauxite.
    lead: :class:`float`
        The price of lead.
    gasoline: :class:`float`
        The price of gasoline.
    munitions: :class:`float`
        The price of munitions.
    steel: :class:`float`
        The price of steel.
    aluminum: :class:`float`
        The price of aluminum.
    food: :class:`float`
        The price of food.
    credits: :class:`float`
        The price of credits.
    """

    id: str
    date: str
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float
    credits: float

    __slots__ = (
        "id",
        "date",
        "coal",
        "oil",
        "uranium",
        "iron",
        "bauxite",
        "lead",
        "gasoline",
        "munitions",
        "steel",
        "aluminum",
        "food",
        "credits",
    )


class Treasure(Data):
    """Represents a treasure.

    Attributes
    ----------
    name: :class:`str`
        The name of the treasure.
    color: :class:`str`
        The treasure's color.
    continent: :class:`str`
        The treasure's continent.
    bonus: :class:`int`
        The treasure's bonus.
    spawndate: :class:`str`
        The treasure's spawndate.
    nation: :class:`Nation`
        The nation that has the treasure.
    """

    name: str
    color: str
    continent: str
    bonus: int
    spawndate: str
    nation: Nation

    __slots__ = ("name", "color", "continent", "bonus", "spawndate", "nation")

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {self.name if 'name' in self.__dict__ else -1}"


class War(Data):
    """Represents a war.

    Attributes
    ----------
    id: :class:`str`
        The war ID.
    date: :class:`str`
        The date the war was declared.
    reason: :class:`str`
        The war reason.
    war_type: :class:`str`
        The war type.
    groundcontrol: :class:`str`
        The ID of the nation that has Ground Control.
    airsuperiority: :class:`str`
        The ID of the nation that has Air Superiority.
    navalblockade: :class:`str`
        The ID of the nation that has a Naval Blockade.
    winner: :class:`str`
        The ID of the winner of the war.
    attacks: Sequence[:class:`WarAttack`]
        The attacks in the war.
    turnsleft: :class:`int`
        The turns left in the war.
    attid: :class:`str`
        The attacker ID.
    att_alliance_id: :class:`str`
        The attacker's alliance ID.
    attacker: :class:`Nation`
        The attacker's nation.
    defid: :class:`str`
        The defender's ID.
    def_alliance_id: :class:`str`
        The defender's alliance ID.
    defender: :class:`Nation`
        The defender's nation.
    attpoints: :class:`int`
        The attacker's Military Action Points.
    defpoints: :class:`int`
        The defender's Military Action Points.
    attpeace: :class:`bool`
        Whether the attacker has offered peace or not.
    defpeace: :class:`bool`
        Whether the defender has offered peace or not.
    att_resistance: :class:`int`
        The attacker's resistance.
    def_resistance: :class:`int`
        The defender's resistance.
    att_fortify: :class:`bool`
        Whether the attacker has fortified or not.
    def_fortify: :class:`bool`
        Whether the defender has fortified or not.
    att_gas_used: :class:`float`
        The amount of gasoline used by the attacker.
    def_gas_used: :class:`float`
        The amount of gasoline used by the defender.
    att_mun_used: :class:`float`
        The amount of munitions used by the attacker.
    def_mun_used: :class:`float`
        The amount of munitions used by the defender.
    att_alum_used: :class:`int`
        The amount of aluminum used by the attacker.
    def_alum_used: :class:`int`
        The amount of aluminum used by the defender.
    att_steel_used: :class:`int`
        The amount of steel used by the attacker.
    def_steel_used: :class:`int`
        The amount of steel used by the defender.
    att_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the attacker.
    def_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the defender.
    att_money_looted: :class:`float`
        The amount of money looted by the attacker.
    def_money_looted: :class:`float`
        The amount of money looted by the attacker.
    att_soldiers_killed: :class:`int`
        The amount of soldiers killed by the attacker.
    def_soldiers_killed: :class:`int`
        The amount of soldiers killed by the defender.
    att_tanks_killed: :class:`int`
        The amount of tanks killed by the attacker.
    def_tanks_killed: :class:`int`
        The amount of tanks killed by the defender.
    att_aircraft_killed: :class:`int`
        The amount of aircraft killed by the attacker.
    def_aircraft_killed: :class:`int`
        The amount of aircraft killed by the defender.
    att_ships_killed: :class:`int`
        The amount of ships killed by the attacker.
    def_ships_killed: :class:`int`
        The amount of ships killed by the defender.
    att_missiles_used: :class:`int`
        The amount of missiles used by the attacker.
    def_missiles_used: :class:`int`
        The amount of missiles killed by the defender.
    att_nukes_used: :class:`int`
        The amount of nukes used by the attacker.
    def_nukes_used: :class:`int`
        The amount of nukes killed by the defender.
    att_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the attacker.
    def_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the defender.
    """

    id: str
    date: str
    reason: str
    war_type: str
    groundcontrol: str
    airsuperiority: str
    navalblockade: str
    winner: str
    attacks: Sequence[WarAttack]
    turnsleft: int
    attid: str
    att_alliance_id: str
    attacker: Nation
    defid: str
    def_alliance_id: str
    defender: Nation
    attpoints: int
    defpoints: int
    attpeace: bool
    defpeace: bool
    att_resistance: int
    def_resistance: int
    att_fortify: bool
    def_fortify: bool
    att_gas_used: float
    def_gas_used: float
    att_mun_used: float
    def_mun_used: float
    att_alum_used: int
    def_alum_used: int
    att_steel_used: int
    def_steel_used: int
    att_infra_destroyed: float
    def_infra_destroyed: float
    att_money_looted: float
    def_money_looted: float
    att_soldiers_killed: int
    def_soldiers_killed: int
    att_tanks_killed: int
    def_tanks_killed: int
    att_aircraft_killed: int
    def_aircraft_killed: int
    att_ships_killed: int
    def_ships_killed: int
    att_missiles_used: int
    def_missiles_used: int
    att_nukes_used: int
    def_nukes_used: int
    att_infra_destroyed_value: float
    def_infra_destroyed_value: float

    __slots__ = (
        "id",
        "date",
        "reason",
        "war_type",
        "groundcontrol",
        "airsuperiority",
        "navalblockade",
        "winner",
        "attacks",
        "turnsleft",
        "attid",
        "att_alliance_id",
        "attacker",
        "defid",
        "def_alliance_id",
        "defender",
        "attpoints",
        "defpoints",
        "attpeace",
        "defpeace",
        "att_resistance",
        "def_resistance",
        "att_fortify",
        "def_fortify",
        "att_gas_used",
        "def_gas_used",
        "att_mun_used",
        "def_mun_used",
        "att_alum_used",
        "def_alum_used",
        "att_steel_used",
        "def_steel_used",
        "att_infra_destroyed",
        "def_infra_destroyed",
        "att_money_looted",
        "def_money_looted",
        "att_soldiers_killed",
        "def_soldiers_killed",
        "att_tanks_killed",
        "def_tanks_killed",
        "att_aircraft_killed",
        "def_aircraft_killed",
        "att_ships_killed",
        "def_ships_killed",
        "att_missiles_used",
        "def_missiles_used",
        "att_nukes_used",
        "def_nukes_used",
        "att_infra_destroyed_value",
        "def_infra_destroyed_value",
    )


class WarAttack(Data):
    """Represents a war attack.

    Attributes
    ----------
    id: :class:`str`
        The attack ID.
    date: :class:`str`
        The date the attack took place.
    type: :class:`str`
        The attack type.
    victor: :class:`str`
        The victor of the attack.
    success: :class:`int`
        The attack's level of success.
    attcas1: :class:`int`
        A special value depending on the attack.
    defcas1: :class:`int`
        A special value depending on the attack.
    attcas2: :class:`int`
        A special value depending on the attack.
    defcas2: :class:`int`
        A special value depending on the attack.
    cityid: :class:`str`
        The ID of the city that was attacked.
    infradestroyed: :class:`float`
        The amount of infrastructure destroyed.
    improvementslost: :class:`int`
        The amount of improvements destroyed.
    moneystolen: :class:`float`
        The amount of money stolen.
    loot_info: :class:`str`
        The loot info of the attack.
    resistance_eliminated: :class:`int`
        The amount of resistance eliminated.
    city_infra_before: :class:`float`
        The city's infrastructure before the attack
    infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed.
    att_mun_used: :class:`float`
        The amount of munitions used by the attacker.
    def_mun_used: :class:`float`
        The amount of munitions used by the defender.
    att_gas_used: :class:`float`
        The amount of gasoline used by the attacker.
    def_gas_used: :class:`float`
        The amount of gasoline used by the defender.
    """

    id: str
    date: str
    type: str
    victor: str
    success: int
    attcas1: int
    defcas1: int
    attcas2: int
    defcas2: int
    cityid: str
    infradestroyed: float
    improvementslost: int
    moneystolen: float
    loot_info: str
    resistance_eliminated: int
    city_infra_before: float
    infra_destroyed_value: float
    att_mun_used: float
    def_mun_used: float
    att_gas_used: float
    def_gas_used: float

    __slots__ = (
        "id",
        "date",
        "type",
        "victor",
        "success",
        "attcas1",
        "defcas1",
        "attcas2",
        "defcas2",
        "cityid",
        "infradestroyed",
        "improvementslost",
        "moneystolen",
        "loot_info",
        "resistance_eliminated",
        "city_infra_before",
        "infra_destroyed_value",
        "att_mun_used",
        "def_mun_used",
        "att_gas_used",
        "def_gas_used",
    )


class PaginatorInfo(Data):
    """Represents paginator info

    Attributes
    ----------
    count: :class:`int`
        Count of items available on the page.
    currentPage: :class:`int`
        The current page number.
    firstItem: :class:`int`
        The index of the first item on the page.
    hasMorePages: :class:`bool`
        Whether there are more pages or not.
    lastItem: :class:`int`
        The index of the last item on the page.
    lastPage: :class:`int`
        The page number of the last page.
    perPage: :class:`int`
        The number of items per page.
    total: :class:`int`
        The total number of items.
    """

    count: int
    currentPage: int
    firstItem: int
    hasMorePages: bool
    lastItem: int
    lastPage: int
    perPage: int
    total: int

    __slots__ = (
        "count",
        "currentPage",
        "firstItem",
        "hasMorePages",
        "lastItem",
        "lastPage",
        "perPage",
        "total",
    )


_ATTRIBUTE_MAP: Mapping[str, Type[Data]] = {
    "alliance": Alliance,
    "cities": City,
    "treasures": Treasure,
    "sent_bankrecs": Bankrec,
    "received_bankrecs": Bankrec,
    "attacks": WarAttack,
    "attacker": Nation,
    "defender": Nation,
    "sender": Nation,
    "receiver": Nation,
    "nations": Nation,
    "nation": Nation,
    "bankrecs": Bankrec,
    "taxrecs": Bankrec,
    "offensive_wars": War,
    "defensive_wars": War,
}



from __future__ import annotations

from typing import TYPE_CHECKING

from .async_ import AsyncData
from .sync import SyncData

if TYPE_CHECKING:
    from csv import _reader  # type: ignore
    from typing import Any, Dict, Sequence, Tuple, Type

    from .data import DumpData

def request_as_bytes(cls: Type[SyncData], type_: str, date: str) -> bytes: ...
def request_as_str(cls: Type[SyncData], type_: str, date: str) -> str: ...
def request_as_csv(cls: Type[SyncData], type_: str, date: str) -> _reader: ...
def request_as_dicts(
    cls: Type[SyncData], type_: str, date: str
) -> Sequence[Dict[str, Any]]: ...
def request_as_dict(
    cls: Type[SyncData], type_: str, date: str
) -> Dict[str, Dict[str, Any]]: ...
def request_as_data(
    cls: Type[SyncData], type_: str, date: str
) -> Tuple[DumpData, ...]: ...
def request_as_data_dict(
    cls: Type[SyncData], type_: str, date: str
) -> Dict[str, DumpData]: ...
async def async_request_as_bytes(
    cls: Type[AsyncData], type_: str, date: str
) -> bytes: ...
async def async_request_as_str(cls: Type[AsyncData], type_: str, date: str) -> str: ...
async def async_request_as_csv(
    cls: Type[AsyncData], type_: str, date: str
) -> _reader: ...
async def async_request_as_dicts(
    cls: Type[AsyncData], type_: str, date: str
) -> Sequence[Dict[str, Any]]: ...
async def async_request_as_dict(
    cls: Type[AsyncData], type_: str, date: str
) -> Dict[str, Dict[str, Any]]: ...
async def async_request_as_data(
    cls: Type[AsyncData], type_: str, date: str
) -> Tuple[DumpData, ...]: ...
async def async_request_as_data_dict(
    cls: Type[AsyncData], type_: str, date: str
) -> Dict[str, DumpData]: ...



from __future__ import annotations

import re
from typing import TYPE_CHECKING, List, Literal, Optional, Union

import requests
from bs4 import BeautifulSoup

__all__ = (
    "scrape_discord_username",
    "alliance_bank_withdraw",
    "scrape_treaties",
    "scrape_treaty_web",
)

if TYPE_CHECKING:
    from typing import TypedDict

    class TreatyData(TypedDict):
        from_: int
        to_: int
        treaty_type: str


def scrape_discord_username(nation_id: int, /) -> Optional[str]:
    """Scrape a nation page for the discord username

    Parameters
    ----------
    nation_id : int
        The nation ID to scrape.

    Returns
    -------
    Optional[str]
        The discord username, or None if not found.
    """
    try:
        response = requests.request(
            "GET", f"https://politicsandwar.com/nation/id={nation_id}"
        )
        return [
            i.contents[1].text  # type: ignore
            for i in BeautifulSoup(response.text, "html.parser").find_all(
                "tr", class_="notranslate"
            )
            if any("Discord Username:" in str(j) for j in i.contents)  # type: ignore
        ][0]
    except IndexError:
        return None


def alliance_bank_withdraw(
    email: str,
    password: str,
    alliance_id: int,
    receiver: str,
    receiver_type: Literal["alliance", "nation"],
    note: Optional[str] = None,
    **resources: Union[int, float, str],
) -> bool:
    """Send money from an alliance bank.

    Parameters
    ----------
    email : str
        The email of the account to use.
    password : str
        The password of the account to use.
    alliance_id : int
        The alliance ID to send from.
    receiver : str
        The receiver of the withdrawal, must be a nation or alliance name.
    receiver_type : Literal["alliance", "nation"]
        The type of receiver, either "alliance" or "nation".
    note : Optional[str], optional
        The note to send with the withdrawal, by default no note is sent.
    **resources : Union[int, float, str]
        The resources to send, specified as kwargs. (i.e. money=100)

    Returns
    -------
    bool
        Whether or not the withdrawal was successful.
    """
    with requests.Session() as session:
        transaction_data = {f"with{key}": value for key, value in resources.items()}
        transaction_data["withtype"] = receiver_type.capitalize()
        if note is not None:
            transaction_data["withnote"] = note
        transaction_data["withrecipient"] = receiver
        transaction_data["withsubmit"] = "Withdraw"
        login_data = {
            "email": email,
            "password": password,
            "loginform": "Login",
        }
        response = session.request(
            "POST", "https://politicsandwar.com/login/", data=login_data
        )
        if "login failure" in response.text.lower():
            return False
        response = session.request(
            "POST",
            f"https://politicsandwar.com/alliance/id={alliance_id}&display=bank",
            data=transaction_data,
        )
        content = response.text
        if "Something went wrong" in content:
            transaction_data["token"] = BeautifulSoup(content, "html.parser").find("input", {"name": "token"}).attrs["value"]  # type: ignore
            response = session.request(
                "POST",
                f"https://politicsandwar.com/alliance/id={alliance_id}&display=bank",
                data=transaction_data,
            )
            content = response.text
        return "successfully transferred" in content


def scrape_treaties(alliance_id: int, /) -> List[TreatyData]:
    """Scrape the treaties of an alliance.

    Parameters
    ----------
    alliance_id : int
        The alliance ID of the alliance to scrape

    Returns
    -------
    List[TreatyData]
        A list of treaties, each treaty is a dict with the keys "from_", "to_", and "treaty_type".
    """
    response = requests.request(
        "GET", f"https://politicsandwar.com/alliance/id={alliance_id}"
    )
    text = response.text
    matches = re.findall(
        r"'from':(\d*), 'to':(\d*), 'color':'\#[\d|\w]*', 'length':\d*, 'title':'(\w*)'",
        text,
    )
    return [
        {"from_": int(i[0]), "to_": int(i[1]), "treaty_type": i[2]} for i in matches
    ]


def scrape_treaty_web() -> List[TreatyData]:
    """Scrape the treaty web

    Returns
    -------
    List[TreatyData]
        A list of treaties, each treaty is a dict with the keys "from_", "to_", and "treaty_type".
    """
    response = requests.request(
        "GET", "https://politicsandwar.com/alliances/treatyweb/all"
    )
    text = response.text
    matches = re.findall(
        r"'from':(\d*), 'to':(\d*), 'color':'\#[\d|\w]*', 'length':\d*, 'title':'(\w*)'",
        text,
    )
    return [
        {"from_": int(i[0]), "to_": int(i[1]), "treaty_type": i[2]} for i in matches
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, cast, overload

import aiohttp

from .base import KitBase
from .data import (
    Alliance,
    Color,
    Data,
    Nation,
    PaginatorInfo,
    Trade,
    Tradeprice,
    Treasure,
    War,
)
from .errors import GraphQLError
from .paginator import AlliancePaginator, NationPaginator, Paginator

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        Literal,
        Mapping,
        MutableMapping,
        Optional,
        Sequence,
        Tuple,
        Type,
        Union,
    )


class AsyncKit(KitBase):
    async def _query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        args: Sequence[Union[str, Any]],
        *,
        is_paginator: bool = False,
    ) -> Dict[str, Any]:
        query = query = self._format_query(endpoint, params, args, is_paginator)
        async with aiohttp.request(
            "GET", self.graphql_url(), json={"query": query}
        ) as response:
            data: Any = await response.json()
            try:
                if "errors" in data[0]:
                    error = (
                        "\n".join(i["message"] for i in data[0]["errors"])
                        if len(data[0]["errors"]) > 1
                        else data[0]["errors"][0]["message"]
                    )
                    raise GraphQLError(error)
            except KeyError:
                pass
            try:
                if "errors" in data:
                    error = (
                        "\n".join(i["message"] for i in data["errors"])
                        if len(data["errors"]) > 1
                        else data["errors"][0]["message"]
                    )
                    raise GraphQLError(error)
            except KeyError:
                pass
            return data

    async def _data_query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        is_paginator: bool = False,
        type_: Type[Data],
        paginator_type: Optional[Type[Paginator]] = None,
        **kwargs: Any,
    ) -> Union[Tuple[Data, ...], Paginator]:
        args = (arg, *args)
        params = params or kwargs
        if "first" not in params and endpoint in {"alliance", "nations"}:
            params["first"] = 5
        response = await self._query(endpoint, params, args, is_paginator=is_paginator)
        if is_paginator and paginator_type:
            data = tuple(type_(i) for i in response["data"][endpoint]["data"])
            if paginator:
                return paginator_type(
                    data, PaginatorInfo(response["data"][endpoint]["paginatorInfo"])
                )
            return data
        data = tuple(type_(i) for i in response["data"][endpoint])
        return data

    @overload
    async def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = False,
        **kwargs: Any,
    ) -> Tuple[Alliance, ...]:
        ...

    @overload
    async def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = True,
        **kwargs: Any,
    ) -> AlliancePaginator:
        ...

    async def alliance_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **kwargs: Any,
    ) -> Union[Tuple[Alliance, ...], AlliancePaginator]:
        data = await self._data_query(
            "alliances",
            params,
            arg,
            *args,
            **kwargs,
            type_=Alliance,
            paginator=paginator,
            is_paginator=True,
            paginator_type=AlliancePaginator,
        )
        if TYPE_CHECKING:
            if isinstance(data, tuple):
                data = cast(Tuple[Alliance, ...], data)
            else:
                data = cast(AlliancePaginator, data)
        return data

    async def color_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Color, ...]:
        data = await self._data_query(
            "colors",
            params,
            arg,
            *args,
            **kwargs,
            type_=Color,
        )
        if TYPE_CHECKING:
            data = cast(Tuple[Color, ...], data)
        return data

    @overload
    async def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = False,
        **kwargs: Any,
    ) -> Tuple[Nation, ...]:
        ...

    @overload
    async def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = True,
        **kwargs: Any,
    ) -> NationPaginator:
        ...

    async def nation_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **kwargs: Any,
    ) -> Union[Tuple[Nation, ...], NationPaginator]:
        data = await self._data_query(
            "nations",
            params,
            arg,
            *args,
            **kwargs,
            type_=Nation,
            paginator=paginator,
            is_paginator=True,
            paginator_type=NationPaginator,
        )
        if TYPE_CHECKING:
            if isinstance(data, tuple):
                data = cast(Tuple[Nation, ...], data)
            else:
                data = cast(NationPaginator, data)
        return data

    async def trade_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Trade, ...]:
        data = await self._data_query(
            "trades",
            params,
            arg,
            *args,
            **kwargs,
            type_=Trade,
        )
        if TYPE_CHECKING:
            data = cast(Tuple[Trade, ...], data)
        return data

    async def trade_price_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Tradeprice, ...]:
        data = await self._data_query(
            "tradeprices",
            params,
            arg,
            *args,
            **kwargs,
            type_=Tradeprice,
        )
        if TYPE_CHECKING:
            data = cast(Tuple[Tradeprice, ...], data)
        return data

    async def treasure_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[Treasure, ...]:
        data = await self._data_query(
            "treasures",
            params,
            arg,
            *args,
            **kwargs,
            type_=Treasure,
        )
        if TYPE_CHECKING:
            data = cast(Tuple[Treasure, ...], data)
        return data

    async def war_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
        **kwargs: Any,
    ) -> Tuple[War, ...]:
        data = await self._data_query(
            "wars",
            params,
            arg,
            *args,
            **kwargs,
            type_=War,
        )
        if TYPE_CHECKING:
            data = cast(Tuple[War, ...], data)
        return data



from __future__ import annotations

from abc import ABCMeta, abstractclassmethod
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple
from zipfile import ZipFile

from .data import DumpData

if TYPE_CHECKING:
    from csv import _reader


class DataBase(metaclass=ABCMeta):
    @staticmethod
    def _validate_params(type_: str, date: str) -> Tuple[str, str]:
        type_ = type_.lower()
        if type_ not in {"alliances", "cities", "nations", "trades", "wars"}:
            raise ValueError(f"Invalid type: {type_}")
        if len(date) != 10:
            raise ValueError(f"Invalid date: {date}")
        if (
            not all(i.isdigit() for i in date[:4] + date[5:7] + date[8:])
            and date[4] + date[7] == "--"
        ):
            raise ValueError(f"Invalid date: {date}")
        return type_, date

    @staticmethod
    def _parse_response_to_bytes(name: str, response: bytes) -> bytes:
        io = BytesIO(response)
        return ZipFile(io).read(name)

    @abstractclassmethod
    def _request(self, type_: str, date: str) -> bytes:
        ...

    @abstractclassmethod
    def request_as_bytes(self, type_: str, date: str) -> bytes:
        """
        Returns the requested CSV as bytes.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts "alliances", "cities", "nations", "trades", or "wars".
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        bytes
            The bytes of the requested CSV.
        """
        ...

    @abstractclassmethod
    def request_as_str(self, type_: str, date: str) -> str:
        """
        Returns the requested CSV as a string.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts ``alliances``, ``cities``, ``nations``, ``trades``, or ``wars``.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        str
            The string of the requested CSV.
        """
        ...

    @abstractclassmethod
    def request_as_csv(self, type_: str, date: str) -> _reader:
        """
        Returns the requested CSV as a :class:`csv._reader`.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts `alliances`, `cities`, `nations`, `trades`, or `wars`.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        csv._reader
            The CSV reader of the requested CSV.
        """
        ...

    @abstractclassmethod
    def request_as_dicts(self, type_: str, date: str) -> Sequence[Dict[str, Any]]:
        """
        Returns the requested CSV as a list of :class:`dict`.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts `alliances`, `cities`, `nations`, `trades`, or `wars`.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        Sequence[Dict[str, Any]]
            The list of dicts of the requested CSV.
        """
        ...

    @abstractclassmethod
    def request_as_dict(self, type_: str, date: str) -> Dict[str, Dict[str, Any]]:
        """
        Returns the requested CSV as a :class:`dict` indexed by ID.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts `alliances`, `cities`, `nations`, `trades`, or `wars`.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        Dict[str, Any]
            The dict of the requested CSV.
        """
        ...

    @abstractclassmethod
    def request_as_data(self, type_: str, date: str) -> Tuple[DumpData, ...]:
        """
        Returns the requested CSV as a tuple of :class:`pnw_dumps.data.DumpData` corresponding to the type of data requested.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts `alliances`, `cities`, `nations`, `trades`, or `wars`.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        Tuple[DumpData, ...]
            The tuple of :class:`pnw_dumps.data.DumpData` corresponding to the type of data requested.
        """
        ...

    @abstractclassmethod
    def request_as_data_dict(self, type_: str, date: str) -> Dict[str, DumpData]:
        """
        Returns the requested CSV as a tuple of :class:`pnw_dumps.data.DumpData` corresponding to the type of data requested indexed by ID.

        Parameters
        ----------
        type_ : str
            The type of data to request. Accepts `alliances`, `cities`, `nations`, `trades`, or `wars`.
        date : str
            The date of the data to request. Accepts the format ``YYYY-MM-DD``.

        Returns
        -------
        Dict[str, DumpData]
            The dict of :class:`pnw_dumps.data.DumpData` corresponding to the type of data requested indexed by ID.
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from .api_key import set_key  # type: ignore
from .core import Kit, async_pnwkit, pnwkit  # type: ignore

if TYPE_CHECKING:
    from typing import Any, Literal, Mapping, MutableMapping, Tuple, Union

    from .async_ import AsyncKit
    from .data import *
    from .paginator import *
    from .sync import SyncKit

__version__ = "1.3.0"

# shortcuts for pnwkit.xxx syntax as opposed to pnwkit.pnwkit.xxx

@overload
async def async_alliance_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **kwargs: Any,
) -> Tuple[Alliance, ...]: ...
@overload
async def async_alliance_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **kwargs: Any,
) -> AlliancePaginator: ...
async def async_alliance_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **kwargs: Any,
) -> Union[Tuple[Alliance, ...], AlliancePaginator]: ...
async def async_color_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Color, ...]: ...
@overload
async def async_nation_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **kwargs: Any,
) -> Tuple[Nation, ...]: ...
@overload
async def async_nation_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **kwargs: Any,
) -> NationPaginator: ...
async def async_nation_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **kwargs: Any,
) -> Union[Tuple[Nation, ...], NationPaginator]: ...
async def async_trade_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Trade, ...]: ...
async def async_trade_price_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Tradeprice, ...]: ...
async def async_treasure_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Treasure, ...]: ...
async def async_war_query(
    self: AsyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[War, ...]: ...
@overload
def alliance_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **kwargs: Any,
) -> Tuple[Alliance, ...]: ...
@overload
def alliance_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **kwargs: Any,
) -> AlliancePaginator: ...
def alliance_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **kwargs: Any,
) -> Union[Tuple[Alliance, ...], AlliancePaginator]: ...
def color_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Color, ...]: ...
@overload
def nation_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **kwargs: Any,
) -> Tuple[Nation, ...]: ...
@overload
def nation_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **kwargs: Any,
) -> NationPaginator: ...
def nation_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **kwargs: Any,
) -> Union[Tuple[Nation, ...], NationPaginator]: ...
def trade_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Trade, ...]: ...
def trade_price_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Tradeprice, ...]: ...
def treasure_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[Treasure, ...]: ...
def war_query(
    self: SyncKit,
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Tuple[War, ...]: ...



from .async_ import AsyncData
from .sync import SyncData

request_as_bytes = SyncData.request_as_bytes
request_as_str = SyncData.request_as_str
request_as_csv = SyncData.request_as_csv
request_as_dicts = SyncData.request_as_dicts
request_as_dict = SyncData.request_as_dict
request_as_data = SyncData.request_as_data
request_as_data_dict = SyncData.request_as_data_dict

async_request_as_bytes = AsyncData.request_as_bytes
async_request_as_str = AsyncData.request_as_str
async_request_as_csv = AsyncData.request_as_csv
async_request_as_dicts = AsyncData.request_as_dicts
async_request_as_dict = AsyncData.request_as_dict
async_request_as_data = AsyncData.request_as_data
async_request_as_data_dict = AsyncData.request_as_data_dict



from setuptools import setup

requirements = ["aiohttp>=3.7.0,<3.8.", "requests>=2.25.0,<2.26.0"]

version = ""
with open("pnwkit/__init__.py") as f:
    try:
        text = f.read()
        index = text.index('__version__ = "') + 15
        text = text[index:]
        index = text.index('"\n')
        version = text[:index]
    except ValueError:
        raise RuntimeError("Version not found")
long_description_content_type = "text/markdown"
readme = ""
with open("README.md") as f:
    readme = f.read()

extras_require = {
    "docs": [
        "sphinx==4.0.3",
        "pydata-sphinx-theme==0.6.3",
    ],
}
packages = ["pnwkit", "pnwkit.ext.dumps", "pnwkit.ext.scrape"]

setup(
    name="pnwkit-py",
    author="Village",
    url="https://github.com/Village05/pnwkit-py",
    project_urls={
        "Documentation": "https://pnwkit-py.readthedocs.io/en/latest/",
        "Issue tracker": "https://github.com/Village05/pnwkit-py/issues",
    },
    version=version,
    packages=packages,
    license="MIT",
    description="A Python wrapper for the Politics and War API.",
    long_description=readme,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires=">=3.8.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    long_description_content_type=long_description_content_type,
    package_data={
        "pnwkit": ["py.typed", "*.pyi"],
        "pnwkit.ext.dumps": ["*.pyi"],
        "pnwkit.ext.scrape": ["*.pyi"],
    },
)
